@@ -21,6 +21,7 @@ use crate::{
     messages,
     request_table::ResponseToken,
     routers::Singleton,
+    tls,
 };
 
 use self::source::Combined;
@@ -87,7 +88,7 @@ impl<C, K, S> Context<C, K, S> {
 
     pub async fn send<M: Message>(&self, message: M) -> Result<(), SendError<M>> {
         let kind = MessageKind::Regular { sender: self.addr };
-        trace!(?message, ">");
+        trace!("> {:?}", message);
         self.do_send(message, kind).await
     }
 
@@ -148,7 +149,7 @@ impl<C, K, S> Context<C, K, S> {
     ) -> Result<(), SendError<M>> {
         let entry = self.book.get_owned(recipient);
         let object = ward!(entry, return Err(SendError(message)));
-        trace!(?message, to = %recipient, ">");
+        trace!(to = %recipient, "> {:?}", message);
         let envelope = Envelope::new(message, MessageKind::Regular { sender: self.addr });
         let fut = object.send(self, envelope.upcast());
         let result = fut.await;
@@ -162,7 +163,7 @@ impl<C, K, S> Context<C, K, S> {
     ) -> Result<(), TrySendError<M>> {
         let entry = self.book.get_owned(recipient);
         let object = ward!(entry, return Err(TrySendError::Closed(message)));
-        trace!(?message, to = %recipient, ">");
+        trace!(to = %recipient, "> {:?}", message);
         let envelope = Envelope::new(message, MessageKind::Regular { sender: self.addr });
 
         object.try_send(envelope.upcast()).map_err(|err| match err {
@@ -182,7 +183,7 @@ impl<C, K, S> Context<C, K, S> {
 
         let message = R::Wrapper::from(message);
         let sender = token.sender;
-        trace!(?message, to = %sender, ">");
+        trace!(to = %sender, "> {:?}", message);
         let envelope = Envelope::new(message, MessageKind::Regular { sender }).upcast();
         let object = ward!(self.book.get(token.sender));
         let actor = ward!(object.as_actor());
@@ -208,6 +209,7 @@ impl<C, K, S> Context<C, K, S> {
         pin_mut!(mailbox_fut);
         pin_mut!(source_fut);
 
+        // TODO: reset trace_id for these logs?
         let envelope = select_biased! {
             envelope = mailbox_fut => match envelope {
                 Some(envelope) => envelope,
@@ -226,6 +228,8 @@ impl<C, K, S> Context<C, K, S> {
             },
         };
 
+        tls::set_trace_id(envelope.trace_id());
+
         let envelope = msg!(match envelope {
             (messages::UpdateConfig { config }, token) => {
                 self.config = config.get().cloned().expect("must be decoded");
@@ -239,7 +243,7 @@ impl<C, K, S> Context<C, K, S> {
             envelope => envelope,
         });
 
-        trace!(message = ?envelope.message(), "<");
+        trace!("< {:?}", envelope.message());
 
         Some(envelope)
     }
@@ -255,12 +259,15 @@ impl<C, K, S> Context<C, K, S> {
             Ok(envelope) => envelope,
             Err(err) => {
                 if err.is_closed() {
+                    // TODO: reset trace_id for this log?
                     trace!("mailbox closed");
                 }
                 return Err(err);
             }
         };
 
+        tls::set_trace_id(envelope.trace_id());
+
         // TODO: poll the sources.
 
         let envelope = msg!(match envelope {
@@ -276,7 +283,7 @@ impl<C, K, S> Context<C, K, S> {
             envelope => envelope,
         });
 
-        trace!(message = ?envelope.message(), "<");
+        trace!("< {:?}", envelope.message());
 
         Ok(envelope)
     }
